#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <pcap.h>

// Default interface.  Assume first network interface is copper ethernet and second
// is wireless.  For OS X, it's "en1" and for Linux it's "eth1."
#ifdef __APPLE__
#define DEFAULT_INTERFACE "en1"
#else
#define DEFAULT_INTERFACE "eth1"
#endif

/** The BPF filter we're applying to the pcap library, so that we only get notified
 * of what we're interested in. */
#define PCAP_FILTER "tcp dst port 80"

/** Ethernet packet headers are always 14 bytes */
#define SIZE_ETHERNET 14

/** Does the user want to quit? */
int userQuit = 0;

/** Persistent buffer in which to store the last capture time as a human-friendly string */
char timeBuffer[40];

/** List of HTTP v1.1 verbs, including WebDav (first three are standard, remainder are dav) */
char verbList[][10] = {
    "GET ",
    "POST ",
    "HEAD ",
    "PUT ",
    "DELETE ",
    "TRACE ",
    "OPTIONS ",
    "SUBSCRIBE ",
    "COPY ",
    "MOVE ",
    "LOCK ",
    "UNLOCK ",
    "MKCOL ",
    "PROPFIND ",
    "PROPPATCH"
};

/** Length of verb list... option can override this if we're just interested in the non-dav stuff */
unsigned int verbListLength = sizeof(verbList) / sizeof(verbList[1]);

unsigned int verbosity = 0;

void signalHandler(int signal) {
    userQuit = 1;
}

void levelPrint(unsigned int level, const char *format, ...) {
    va_list ap;
    if (level > verbosity)
        return;
    va_start(ap, format);
    vprintf(format, ap);
    va_end(ap);
} 

int printUsage(const char *errorMessage) {
    if (errorMessage && *errorMessage)
        fprintf(stderr, "Error: %s\n", errorMessage);
    fprintf(stderr, 
        "wcap [-v] [-d] [interface]\n"
        "where:\n"
        "'interface' is the ethernet port on which to sniff traffic, but\n"
        " defaults to en1/eth1\n"
        " (example: en0 or en1 on OS X, eth0 or eth1 on Linux)\n"
        "-v increases verbosity (use multiple times for even more info)\n"
        "-d disable webdav verbs like PUT, DELETE, LOCK, etc.\n"
        );
    return 1;
}

/** Take the current time (as seconds since Epoc plus microseconds) and convert to string */
const char *printTime(const struct timeval *tv) {
    char usec[10];
    strftime(timeBuffer, sizeof(timeBuffer), "%Y-%m-%d %H:%M:%S.", localtime(&tv->tv_sec));
    sprintf(usec, "%03d", tv->tv_usec / 1000);
    strcat(timeBuffer, usec);
    return timeBuffer;
}

void printContent(const char *timestamp, const struct in_addr *source, const struct in_addr *destination, const char *payload, int payloadLength) {
    int i;
    char *path;
    char *host = NULL;
    const char *payloadEnd = payload + payloadLength;
    const char *pos = payload;
    const char *start = NULL, *end = NULL;
    // Get path from URL (between the spaces in "GET /images/blah.jpg HTTP/1.1")
    while ((pos < payloadEnd) && (end == NULL)) {
        if (*pos == ' ') {
            if (start == NULL)
                start = pos + 1;
            else
                end = pos;
        }
        pos++;
    }
    if (end == NULL) {
        levelPrint(1, "Malformed HTTP request packet\n");
        return;
    }
    path = (char *) malloc(end - start + 1);
    strncpy(path, start, end - start);
    path[end - start] = 0;
    // Get the host name (if available)
    pos = payload;
    start = end = NULL;
    while ((pos < payloadEnd) && (start == NULL)) {
        if (strncasecmp(pos, "\r\nHost: ", 8) == 0) {
            start = pos + 8; 
        }
        pos++;
    }
    if (start) {
        pos = start+1;
        while ((pos < payloadEnd) && (end == NULL)) {
            if (*pos == '\r')
                end = pos;
            pos++;
        }
        if (end) {
            host = (char *) malloc(end - start + 1);
            strncpy(host, start, end - start);
            host[end - start] = 0;
        } else
            levelPrint(1, "Malformed Host header: no end\n");
    }
    // If not found, fall back to numeric IP address
    if (host == NULL)
        host = strdup(inet_ntoa(*destination));
    // Print packet
    printf("%s\t%s\thttp://%s%s\n", timestamp, inet_ntoa(*source), host, path);
    free(host);
    free(path);
    if (verbosity >= 3) {
        for (i=0; i<payloadLength; i++, payload++)
            if ((*payload > ' ') && (*payload <= '~'))
                printf("%c", *payload);
            else
                printf(" 0x%02X ", *payload);
        printf("\n");
    }
}

void handlePacket(u_char *args, const struct pcap_pkthdr *header, const u_char *packet) {
    const struct ethernet *ethernet;  /// The ethernet header
    const struct ip *ip;              /// The IP header
    const struct tcphdr *tcp;            /// The TCP header
    const char *payload;              /// Packet payload
    u_int ipLength;
    u_int tcpLength;
    u_int payloadLength;
    u_int i;
    unsigned char match = 0; 

    levelPrint(2, "Got packet.\n");
    // Align the structures
    ethernet = (struct ethernet *)packet;
    ip = (struct ip *)(packet + SIZE_ETHERNET);
    ipLength = ip->ip_hl * 4; //IP_HL(ip) * 4;
    if (ipLength < 20) {
        fprintf(stderr, "Warning: Invalid IP header length: %u bytes\n", ipLength);
        return;
    }
    tcp = (struct tcphdr *)(packet + SIZE_ETHERNET + ipLength);
    tcpLength = tcp->th_off * 4; //TH_OFF(tcp)*4;
    if (tcpLength < 20) {
        fprintf(stderr, "Warning: Invalid TCP header length: %u bytes\n", tcpLength);
        return;
    }
    payload = (const char *)(packet + SIZE_ETHERNET + ipLength + tcpLength);
    payloadLength = header->caplen - (SIZE_ETHERNET + ipLength + tcpLength);
    // Check if this is a packet we are interested in
    // NOTE: for the future, this is more efficiently done as a state machine
    for (i=0; (i<verbListLength) && (match == 0); i++)
        if (strncmp(payload, verbList[i], strlen(verbList[i])) == 0)
            match = 1;
    if (!match) {
        levelPrint(1, "Got packet, ignoring: no match for HTTP verb.\n");
        return;
    }
    printContent(printTime(&header->ts), &ip->ip_src, &ip->ip_dst, payload, payloadLength);
}

int sniff(const char *device) {
    char errbuf[PCAP_ERRBUF_SIZE];  /// Error message returned from pcap
    pcap_t *handle;                 /// The handle of the device
    struct bpf_program fp;          /// The compiled BPF filter
    bpf_u_int32 mask;               /// The netmask of our sniffing device
    bpf_u_int32 net;                /// The IP of our sniffing device

    // Get the IP/netmask of our device (needed for pcap_compile)    
    if (pcap_lookupnet(device, &net, &mask, errbuf) == -1) {
        fprintf(stderr, "Warning: Can't get netmask for device %s\n", device);
        net = 0;
        mask = 0;
    }
    // Open the device    
    if ((handle = pcap_open_live(device, BUFSIZ, 1, 1000, errbuf)) == NULL) {
        fprintf(stderr, "Error: Couldn't open device %s, %s\n", device, errbuf);
        return 2;
    }
    // Compile the filter
    if (pcap_compile(handle, &fp, PCAP_FILTER, 0, net) == -1) {
        fprintf(stderr, "Error: Couldn't parse filter %s, %s\n", PCAP_FILTER, pcap_geterr(handle));
        return 3;
    }
    if (pcap_setfilter(handle, &fp) == -1) {
        fprintf(stderr, "Error: Couldn't install filter %s, %s\n", PCAP_FILTER, pcap_geterr(handle));
        return 4;
    }
    if (pcap_loop(handle, -1, handlePacket, NULL) == -1) {
        fprintf(stderr, "Error: Couldn't set callback, %s\n", pcap_geterr(handle));
        return 4;
    }
    return 0;
}

int main(int argc, char ** argv) {
    const char *interface = DEFAULT_INTERFACE;
    int i, rc, ch;
    // Only root can set the interface in promiscuous mode
    if ((getuid() != 0) && (geteuid() != 0))
        return printUsage("This application must be run as root");
    while ((ch = getopt(argc, argv, "vd")) != -1) {
        switch(ch) {
            case 'v':
                verbosity++;
                break;
            case 'd':
                verbListLength = 3;
                break;
            case '?':
            default:
                return printUsage("Unknown flag");
        }
    }
    argc -= optind;
    argv += optind;
    if (argc >= 1)
        interface = argv[0];
    //verbosity = 2;
    printf("Scanning for HTTP verbs: ");
    for (i=0; i<verbListLength; i++)
        printf("%s", verbList[i]);
    printf("\n");
    // Set up signal handler so we can exit cleanly
    //signal(SIGTERM, signalHandler); // kill/killall handler
    //signal(SIGHUP, signalHandler);  // hangup handler
    //signal(SIGINT, signalHandler);  //  ^C handler
    // Do it!
    printf("Listening on network interface %s\n", interface);
    if ((rc = sniff(interface)) != 0)
        return rc;
    printf("Successful termination\n");
    return 0;
}
